import os
import uuid
from datetime import datetime
from typing import Tuple

from ..aws import SQSProducer
from ..chatbot import Chatbot
from ..common import (AccountType, DataEnconding, MessageContentType,
                      MessageType, Platform, WaterfallMode)
from ..email_task import Email
from ..mobile_push import Push
from ..proto import notification_hub_pb2 as pb
from ..sms import Sms
from ..whatsapp import Whatsapp


class Task:
    """
    A wrapper class for NotificationTask protobuf structure
    """

    def __init__(self, name: str,
                       sent_by_id: str,
                       client: str,
                       platform: Platform,
                       message_type: MessageType = MessageType.MARKETING,
                       email: Email = None,
                       sms: Sms = None,
                       whatsapp: Whatsapp = None,
                       push: Push = None,
                       waterfall_type: WaterfallMode = WaterfallMode.AUTO,
                       event_id: str = "",
                       chatbot: Chatbot = None):
        """
        Parameters:
            name (str): A unique name for this notification task
            sent_by_id (str): ID of the user who triggered the notification
            client (str): Name of the application or service triggering the notification
            platform (Platform): Which company vertical are we sending from
            message_type (MessageType, optional): Medium of the notifications (defaults to MessageType.MARKETING)
            email (Email, optional): The Email object (refer Email docs)
            sms (Sms, optional): The SMS object (refer SMS docs)
            whatsapp (Whatsapp, optional): The WhatsApp object  (refer WhatsApp docs)
            push (Push, optional): The Push object (refer Push docs)
            waterfall_type (WaterfallMode, optional): Whether to override the default priority engine logic or not
            event_id (str): event id to connect notifications to specific applications
            chatbot (chatbot, optional): Type of the chatbot (defaults to None)
        """
        self._task = pb.NotificationTask()

        # Check whether at least one channel is specified
        if not email and not sms and not whatsapp and not push:
            raise AssertionError('At least one channel should be passed')
        if not whatsapp and chatbot and chatbot.chatbot_whatsapp is not None:
            raise AssertionError('ChatbotWhatsapp is used without Whatsapp structure')
        self.__set_id()
        self.__set_triggered_on()

        self._task.name = name
        self._task.sentByID = sent_by_id
        self._task.client = client
        self._task.platform = platform
        self._task.messageType = message_type
        self._task.waterfallType = waterfall_type
        self._task.eventID = event_id
        
        # Assign the notification channels
        self.__set_sms(sms)
        self.__set_email(email)
        self.__set_whatsapp(whatsapp)
        self.__set_push(push)
        self.__set_chatbot(chatbot)

    def __set_id(self):
        """
        Generates a UUID v4 and assigns the value as ID
        """
        self._task.ID = str(uuid.uuid4())

    def __set_triggered_on(self):
        """
        Sets the current timestamp as triggered on
        """
        now = datetime.now()
        self._task.triggeredOn = int(datetime.timestamp(now))

    def __set_sms(self, value: Sms):
        if not value:
            return
        self._task.sms.CopyFrom(value.proto)

    def __set_email(self, value: Email):
        if not value:
            return
        # Set the platform
        value.platform = self._task.platform
        self._task.email.CopyFrom(value.proto)

    def __set_whatsapp(self, value: Whatsapp):
        if not value:
            return
        self._task.whatsapp.CopyFrom(value.proto)

    def __set_push(self, value: Push):
        if not value:
            return
        self._task.push.CopyFrom(value.proto)
        
    def __set_chatbot(self, value: Chatbot):
        if not value:
            return
        self._task.chatbot.CopyFrom(value.proto)

    @property
    def proto(self) -> pb.NotificationTask:
        return self._task

    def send(self, **kwargs) -> Tuple[str, str]:
        """
        Sends the notification task to Notification Hub Queue
        :returns: The Task ID and the AWS Message ID.
        """

        if self._task.messageType == MessageType.MARKETING:
            queue_name = os.getenv('NOTIFICATION_HUB_MARKETING_SQS_QUEUE_NAME')
        elif self._task.messageType == MessageType.TRANSACTIONAL:
            queue_name = os.getenv('NOTIFICATION_HUB_SQS_QUEUE_NAME')
        elif self._task.messageType == MessageType.OTP:
            queue_name = os.getenv('NOTIFICATION_HUB_OTP_SQS_QUEUE_NAME')

        kwargs.update({"queue_name": queue_name})

        producer = SQSProducer(**kwargs)
        aws_msg_id = producer.send_message(self._task, self._task.messageType)
        return self._task.ID, aws_msg_id



from .aws import Session, SQSProducer
from .chatbot import Chatbot, ChatbotWhatsapp
from .common import Platform, Waterfall, WaterfallMode
from .email_task import Email, EmailAttachment, EmailRecipient
from .mobile_push import Push
from .sms import Sms
from .task import Task
from .whatsapp import Whatsapp

__all__ = ["Email", "EmailAttachment", "EmailRecipient",
           "Sms", "Whatsapp", "Push", "Task",
           "WaterfallMode", "Waterfall", "Platform", "Session", "SQSProducer", "Chatbot", "ChatbotWhatsapp"]



import re
from datetime import datetime, timedelta

from validate_email import validate_email as ve


class InvalidTemplateURL(Exception):
    pass


class InvalidAttachmentUrl(Exception):
    pass


class InvalidEmail(Exception):
    pass


class InvalidMobile(Exception):
    pass


class InvalidToken(Exception):
    pass


class ImproperlyConfigured(Exception):
    pass


def get_expiry(offset: int) -> int:
    """
    Returns the expiry timestamp in epoch
    :param offset the expiry offset in days
    :return: Unix timestamp in int
    """
    expiry_dt = datetime.utcnow() + timedelta(days=offset)
    expiry_epoch = datetime.timestamp(expiry_dt)
    return int(expiry_epoch)


def validate_template(url: str):
    """
    Validates a URL
    :param url: The value provided
    :raises: InvalidTemplateURL if not a valid url
    """
    regex = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)
    if not re.match(regex, url):
        raise InvalidTemplateURL("Invalid template provided")


def validate_attachment_url(url: str):
    """
    Validates a URL of an email attachment
    :param url: The value provided
    :raises: InvalidAttachmentUrl if not a valid url
    """
    try:
        validate_template(url)
    except InvalidTemplateURL:
        raise InvalidAttachmentUrl


def validate_email(email: str):
    """
    Validates a email address
    :param email: The value provided
    :raises: InvalidEmail if not a valid email address
    """
    if not ve(email):
        raise InvalidEmail("Invalid email '{}' provided".format(email))


def validate_mobile(mobile: str):
    """
    Validates a mobile number
    :param mobile: The mobile no.
    :raises: InvalidMobile if not a valid mobile number
    """
    regex = re.compile(r'^[6-9]\d{9}$')
    if not re.match(regex, mobile):
        raise InvalidMobile('Invalid mobile number provided.')


def validate_arn_endpoint(endpoint: str):
    if endpoint is None or len(endpoint) == 0:
        raise InvalidToken("Invalid token")



from ..common import AccountType, DataEnconding, MessageContentType
from ..proto import notification_hub_pb2 as pb


class ChatbotWhatsapp:
    def __init__(
            self,
            account_type: AccountType = AccountType.HSM,
            message_content_type: MessageContentType = MessageContentType.TEXTTYPE,
            is_hsm: bool = False,
            is_template: bool = False,
            button_url_param: str = "",
            data_encoding: DataEnconding = DataEnconding.TEXTDATA,
            message_id: str ="",
            extra_param: str = ""
    ):
        """
        Parameters:
            account_type (AccountType, optional): account type to be used to send the msg (defaults to AccountType.HSM; other options are TWOWAY)
            message_content_type (MessageContentType, optional): message content type (defaults to MessageContentType.TEXTTYPE)
            is_hsm (bool, optional): for HSM temaplte; (defaults to False)
            is_template (bool, optional): for templated message; (defaults to False)
            button_url_param (str, optional): buttom params to send
            data_encoding (DataEnconding, optional): (defaults to DataEnconding.TEXTDATA; set UNICODETEXTDATA for the Non-English chars or Emojis)
            message_format (MessageFormat, optional): format of the message; (defaults to MessageFormat.JSON; other options are TEXT and XML)
            message_id (str, optional): for a session id
            extra_params (str, optional): additional information to send
        """
        self._chatbot_whatsapp = pb.ChatbotWhatsapp()
        self._chatbot_whatsapp.accountType = account_type
        self._chatbot_whatsapp.messageContentType = message_content_type
        self._chatbot_whatsapp.isHSM = is_hsm
        self._chatbot_whatsapp.isTemplate = is_template
        self._chatbot_whatsapp.buttonUrlParam = button_url_param
        self._chatbot_whatsapp.dataEncoding = data_encoding
        self._chatbot_whatsapp.messageID = message_id
        self._chatbot_whatsapp.extraParam = extra_param

    @property
    def proto(self) -> pb.ChatbotWhatsapp:
        return self._chatbot_whatsapp



import json

from ..common import ChatbotType
from ..proto import notification_hub_pb2 as pb
from .chatbot_whatsapp import ChatbotWhatsapp


class Chatbot:
    def __init__(
            self,
            chatbot_type: ChatbotType = ChatbotType.JOBSEARCH,
            chatbot_whatsapp: ChatbotWhatsapp = None
    ):
        """
        Parameters:
            chatbot_type (ChatbotType): Type of chatbot; defaults to ChatbotType.JOBSEARCH
            chatbot_whatsapp (ChatbotWhatsapp): Additional information related whatsapp 
        """
        self._chatbot = pb.Chatbot()
        self._chatbot.chatbotType = chatbot_type
        self.__set_chatbot_whatsapp(chatbot_whatsapp)

    def __set_chatbot_whatsapp(self, value: ChatbotWhatsapp = None):
        if not value:
            value = ChatbotWhatsapp()
        self._chatbot.chatbotWhatsapp.CopyFrom(value.proto)

    @property
    def proto(self) -> pb.Chatbot:
        return self._chatbot



from os import path

import boto3

from notificationhub_sdk.base import ImproperlyConfigured

TOKEN_PATH = "/var/run/secrets/eks.amazonaws.com/serviceaccount/token"
SQS_SERVICE_NAME = "sqs"

class Session:
    """
    Used to create AWS Session
    """
    
    def __init__(self,
                 region: str,
                 access_key: str,
                 secret_key: str, 
                 eks_role_arn: str):
        """
        Parameters:
            access_key: access key of aws account
            secret_key: secret key of aws account
            eks_role_arn: eks role which has access to SQS 
        """
        self._region = region
        self._acces_key = access_key
        self._secret_key = secret_key
        self._eks_role_arn = eks_role_arn
        
    def get(self):
        """
        creates aws session based on different criteria
        - Access key and secret key given in the settings
        - EC2: based on the assumption that the role is given to the EC2 instance
        - EKS-role based access: eks_role_arn is taken into consideration
        - Local settings: read from the local ~/.aws/credentials path
        """
        # Access key and secret key given in the settings
        if self._acces_key and self._secret_key:
            session = self.__handle_access_key_authentication()
        else:
            session = self.__handle_role_based_authentication()
        
        # local development
        if session is None:
            session = self.__handle_local_dev_authentication()

        # check whether session is created after all possible options
        if session is None:
            raise ImproperlyConfigured('AWS session is not created due to improper setting provided in region `{}`'.format(self._region))

        return session
    
    def __handle_access_key_authentication(self):
        """
        creates an AWS session based on the Access Key and Secret Key
        """
        client_kwargs = {
            'service_name': SQS_SERVICE_NAME,
            'aws_access_key_id': self._acces_key,
            'aws_secret_access_key': self._secret_key,
            'region_name': self._region
        }
        return boto3.resource(**client_kwargs)
    
    def __handle_role_based_authentication(self):
        """
        creates an AWS session based on IAM role given to EC2 or Pod
        """
        # EC2 instance
        client_kwargs = {
            'service_name': SQS_SERVICE_NAME,
            'region_name': self._region
        }
        session = boto3.resource(**client_kwargs)

        # eks role-based access
        if self._eks_role_arn and self._eks_role_arn != "":
            # check whether the given information has the account number or not
            aws_client = boto3.client('sts')
            if aws_client == None or aws_client == "":
                raise ImproperlyConfigured('Issue while connecting to SQS region `{}`'.format(self._region))

            try:
                if not path.exists(TOKEN_PATH):
                    raise ImproperlyConfigured('Token not present in pod. Check token at `{}`'.format(TOKEN_PATH))
                with open(TOKEN_PATH, 'r') as file:
                    access_details = file.read()
                web_identity_response = aws_client.assume_role_with_web_identity(
                    RoleArn=self._eks_role_arn,
                    RoleSessionName='nh_python_sdk_role_session',
                    WebIdentityToken=access_details,
                )
                if not web_identity_response:
                    raise ImproperlyConfigured('Issue with token in pod. Check token at `{0}` with {1} '.format(TOKEN_PATH, self._eks_role_arn))
                
                response_credentials = web_identity_response.get('Credentials')
                if not response_credentials:
                    raise ImproperlyConfigured('Web identity response does not have the Credentials key `{0}` with {1} '.format(TOKEN_PATH, self._eks_role_arn))
                
                # get credentials from the web identity response
                response_access_key = response_credentials.get('AccessKeyId')
                response_secret_key = response_credentials.get('SecretAccessKey')
                response_session_token = response_credentials.get('SessionToken')
                
                if not response_access_key or not response_secret_key or not response_session_token :
                    raise ImproperlyConfigured('Keys retrieved from the web identity response are none `{0}` with {1} '.format(TOKEN_PATH, self._eks_role_arn))
                
                client_kwargs = {
                    'service_name': SQS_SERVICE_NAME,
                    'region_name': self._region,
                    'aws_access_key_id': response_access_key,
                    'aws_secret_access_key': response_secret_key,
                    'aws_session_token': response_session_token
                }
                session = boto3.resource(**client_kwargs)

            except ImportError:
                raise ImproperlyConfigured('Token not present in pod. Check token at `{0}` with {1} '.format(TOKEN_PATH, self._eks_role_arn))
            
        return session

    def __handle_local_dev_authentication(self):
        """
        creates an AWS session from ~/.aws/credentials file
        """
        return boto3.resource(SQS_SERVICE_NAME)
